import signal
import typing as tp

from satella.time import sleep

end = False


def __sighandler(a, b):
    global end
    end = True


def hang_until_sig(extra_signals: tp.Optional[tp.Sequence[int]] = None) -> None:
    """
    Will hang until this process receives SIGTERM or SIGINT.
    If you pass extra signal IDs (signal.SIG*) with extra_signals,
    then also on those signals this call will release.

    :param extra_signals: a list of extra signals to listen to
    """
    global end
    extra_signals = extra_signals or ()

    old_term = signal.getsignal(signal.SIGTERM)
    old_int = signal.getsignal(signal.SIGINT)
    olds = []

    # Set the signal handler
    signal.signal(signal.SIGTERM, __sighandler)
    signal.signal(signal.SIGINT, __sighandler)
    for s in extra_signals:
        olds.append(signal.getsignal(s))
        signal.signal(s, __sighandler)

    while not end:
        sleep(0.5, True)

    # Unset the signal handler
    signal.signal(signal.SIGTERM, old_term)
    signal.signal(signal.SIGINT, old_int)
    for s, old_handler in zip(extra_signals, olds):
        signal.signal(s, old_handler)

    end = False  # reset for next use



def update_if_not_none(dictionary: dict, key, value):
    """
    Syntactic sugar for

    >>> if value is not None:
    >>>     dictionary[key] = value

    :param dictionary: dictionary to update
    :param key: key to use
    :param value: value to use
    :return: the dictionary itself
    """
    if value is not None:
        dictionary[key] = value
    return dictionary



import queue
import typing as tp

from satella.exceptions import PreconditionError

T = tp.TypeVar('T')
U = tp.TypeVar('U')
Expression = tp.NewType('Expression', str)


# noinspection PyPep8Naming
def _TRUE(x):
    return True


Queue = tp.TypeVar('Queue')


def queue_get(queue_getter: tp.Callable[[object], Queue], timeout: tp.Optional[float] = None,
              exception_empty=queue.Empty,
              queue_get_method: tp.Callable[[Queue, tp.Optional[float]], tp.Any] =
              lambda x, timeout: x.get(
                  timeout=timeout)):
    """
    A decorator for class methods that consume from a queue.

    Timeout of None means block forever.

    First attribute of the decorator-given function must be a normal instance method
    accepting an element taken from the queue, so it must accepts two arguments - first is
    self, second is the element from the queue.

    :param queue_getter: a callable that will render us the queue, or a string, which will be
        translated to a property name
    :param timeout: a timeout to wait. If timeout happens, simple no-op will be done and None
        will be returned.
    :param exception_empty: exception (or a tuple of exceptions) that are raised on queue being
        empty.
    :param queue_get_method: a method to invoke on this queue. Accepts two arguments - the first
        is the queue, the second is the timeout. It has to follow the type signature given.

    Use instead of:

    >>> class QueueProcessor:
    >>>     def __init__(self, queue):
    >>>         self.queue = queue
    >>>     def do(self):
    >>>         try:
    >>>             msg = self.queue.get(timeout=TIMEOUT)
    >>>         except queue.Empty:
    >>>             return

    Instead of aforementioned code, please use:

    >>> class QueueProcessor:
    >>>     def __init__(self, queue):
    >>>         self.queue = queue
    >>>     @queue_get(lambda self: self.queue, timeout=TIMEOUT)
    >>>     def do(self, msg):
    >>>         ...
    """
    if isinstance(queue_getter, str):
        my_queue_getter = lambda x: getattr(x, queue_getter)
    else:
        my_queue_getter = queue_getter

    def outer(fun):
        @wraps(fun)
        def inner(self):
            try:
                que = my_queue_getter(self)
                item = queue_get_method(que, timeout)
            except exception_empty:
                return
            return fun(self, item)

        return inner

    return outer


# taken from https://stackoverflow.com/questions/1288498/using-the-same-decorator-with-arguments-wi\
# th-functions-and-methods
def chain_functions(fun_first: tp.Callable[..., tp.Union[tp.Tuple[tp.Tuple, tp.Dict],
                                                         tp.Dict, tp.Tuple]]) -> tp.Callable:
    """
    A decorator to chain function calls.
    This function is expected to return:

     * a 2-tuple [tp.Tuple, tp.Dict] - args and kwargs for the next function
     * tp.Dict - only kwargs will be passed
     * any other tuple - only args will be passed
     * any other type - will be passed as a first argument

    of arguments to pass to wrapped function. So this

    >>> def test3(...):
    >>>     ...
    >>> def test2(...):
    >>>     ...
    >>> def test(...):
    >>>     args, kwargs = test2(...)
    >>>     return test(3)
    >>> v = test(a, b, c)

    Is equivalent to this:
    >>> @chain_functions
    >>> def test2(...):
    >>>     ...
    >>> @test2
    >>> def test3(...):
    >>>     ...
    >>> v = test3(a, b, c)
    """

    def outer(fun_next):
        @wraps(fun_next)
        def inner(*args, **kwargs):
            ret = fun_first(*args, **kwargs)
            if isinstance(ret, dict):
                args = ()
                kwargs = ret
            elif isinstance(ret, tuple) and len(ret) == 2:
                args, kwargs = ret
            elif isinstance(ret, tuple):
                args = ret
                kwargs = {}
            else:
                args = ret,
                kwargs = {}
            return fun_next(*args, **kwargs)

        return inner

    return outer


def wraps(cls_to_wrap: tp.Type) -> tp.Callable[[tp.Type], tp.Type]:
    """
    A functools.wraps() but for classes.

    As a matter of fact, this can replace functools.wraps() entirely.

    :param cls_to_wrap: class to wrap
    """

    def outer(cls: tp.Type) -> tp.Type:
        if hasattr(cls_to_wrap, '__doc__'):
            cls.__doc__ = cls_to_wrap.__doc__
        if hasattr(cls_to_wrap, '__name__'):
            cls.__name__ = cls_to_wrap.__name__
        if hasattr(cls_to_wrap, '__module__'):
            cls.__module__ = cls_to_wrap.__module__
        if hasattr(cls_to_wrap, '__annotations__'):
            cls.__annotations__ = cls_to_wrap.__annotations__
        return cls

    return outer


def has_keys(keys: tp.List[str]):
    """
    A decorator for asserting that a dictionary has given keys. Will raise PreconditionError if
    it doesn't.

    This outputs a callable that accepts a dict and returns True if it has all the keys
    necessary.

    Returns True if the dict has all necessary keys.

    This is meant to be used in conjunction with @precondition

    :param keys: list of keys to expect
    """

    def inner(dictionary: dict) -> bool:
        for key in keys:
            if key not in dictionary:
                raise PreconditionError('Key %s not found in dict' % (key,))
        return True

    return inner


def short_none(clb: tp.Union[Expression, tp.Callable[[T], U]]) -> tp.Callable[
    [tp.Optional[T]], tp.Optional[U]]:
    """
    Accept a callable. Return a callable that executes it only if passed a no-None arg,
    and returns its result. If passed a None, return a None

    callable can also be a string, in this case it will be appended to lambda x: and eval'd

    :param clb: callable/1->1
    :return: a modified callable
    """
    if isinstance(clb, str):
        q = dict(globals())
        exec('_callable = lambda x: ' + clb, q)
        clb = q['_callable']

    @wraps(clb)
    def inner(arg: tp.Optional[T]) -> tp.Optional[U]:
        if arg is None:
            return None
        else:
            return clb(arg)

    return inner



import typing as tp

from satella.exceptions import ConfigurationSchemaError
from satella.imports import import_class
from .base import Descriptor
from .registry import BASE_LOOKUP_TABLE
from .structs import create_key, Dict


def _get_descriptor_for_str(key: str, value: str) -> Descriptor:
    try:
        if value in ('int', 'float', 'str', 'ipv4', 'any', 'bool'):
            return create_key(BASE_LOOKUP_TABLE[value](),
                              key, False, None)
    except KeyError:
        raise ConfigurationSchemaError('Unknown descriptor type %s' % (value, ))


def _get_descriptor_for_dict(key: str, value: dict) -> Descriptor:
    if 'type' not in value:
        return create_key(descriptor_from_dict(value),
                          key, False, None)
    else:
        args = ()
        type_ = value['type']
        if type_ == 'list':
            of = _get_descriptor_for('', value.get('of', ''))
            args = (of,)
        elif type_ == 'caster':
            cast_to = value.get('cast_to')
            dict_to_look_in = globals().copy()
            if not isinstance(__builtins__, dict):
                dict_to_look_in.update(__builtins__.__dict__)
            else:
                dict_to_look_in.update(__builtins__)
            dict_to_look_in.update(locals())
            if cast_to in dict_to_look_in:
                y = dict_to_look_in[cast_to]
            else:
                y = import_class(cast_to)

            if 'expr' in value:
                dict_ = globals().copy()
                dict_.update(locals())
                dict_['y'] = y
                args = eval('lambda x: ' + value['expr'], dict_),
            else:
                args = y,
        elif type_ == 'union':
            args = [_get_descriptor_for('', x) for x in value.get('of', [])]
        optional, default = False, None
        if 'default' in value:
            optional = True
            default = value['default']
        optional = value.get('optional', optional)
        descriptor = BASE_LOOKUP_TABLE[type_](*args)
        return create_key(descriptor, key, optional, default)


def _get_descriptor_for(key: str, value: tp.Any) -> Descriptor:
    if value == '':
        return Descriptor()
    elif isinstance(value, str):
        return _get_descriptor_for_str(key, value)
    elif isinstance(value, dict):
        return _get_descriptor_for_dict(key, value)
    elif isinstance(value, type):
        return _get_descriptor_for(key, value.__qualname__)
    else:
        raise ConfigurationSchemaError('invalid schema, unrecognized config object %s' % (value,))


def descriptor_from_dict(dct: dict) -> Descriptor:
    """
    Giving a Python dictionary-defined schema of the configuration, return a Descriptor-based one

    :param dct: something like
    {
        "a": "int",
        "b": "str",
        "c": {
            "type": "int"
            "optional": True,
            "default": 5
        },
        "d": {
            "a": "int",
            "b": "str"
        }
    }

    although you can pass "int", "float" and "str" without enclosing quotes, that will work too

    :return: a Descriptor-based schema
    """
    fields = []

    for key, value in dct.items():
        descriptor = _get_descriptor_for(key, value)
        fields.append(descriptor)

    return Dict(fields)



import warnings

from .os import daemonize, PIDFileLock, hang_until_sig, suicide, is_running_as_root

warnings.warn('Importing from satella.posix is deprecated, import from satella.os instead')

__all__ = [
    'daemonize',
    'PIDFileLock', 'hang_until_sig',
    'suicide', 'is_running_as_root'
]



import os
import sys
import typing as tp
import warnings

from satella.coding.recast_exceptions import rethrow_as
from satella.exceptions import ConfigurationError
from .base import BaseSource
from .format import JSONSource

__all__ = [
    'EnvVarsSource', 'EnvironmentSource',
]


class EnvironmentSource(BaseSource):
    """
    This just returns a dictionary of { env_name => that env's value }

    :param env_name: name of the environment variable to check for
    :param config_name: name of the env_name in the dictionary to return
    :param cast_to: callable that converts a string to whatever form is desired.
        Note that this is **deprecated**. Use schema instead to cast to desired typing.
    """
    __slots__ = ('env_name', 'config_name', 'cast_to')

    def __init__(self, env_name: str, config_name: tp.Optional[str] = None, cast_to=None):
        super(EnvironmentSource, self).__init__()
        self.env_name = env_name  # type: str
        self.config_name = config_name or env_name  # type: str
        if cast_to is not None:
            warnings.warn(
                'Use .schema module to cast values to target type. '
                'This module is only concerned with loading the configuration', DeprecationWarning)
            self.cast_to = cast_to  # type: tp.Callable[[tp.Any], tp.Any]
        else:
            self.cast_to = lambda x: x

    @rethrow_as((ValueError, TypeError, KeyError), ConfigurationError)
    def provide(self) -> dict:
        v = self.cast_to(os.environ[self.env_name])

        return {self.config_name: v}


class EnvVarsSource(JSONSource):
    """
    Return a dictionary that is the JSON encoded within a particular environment variable
    """
    __slots__ = ('env_name',)

    def __init__(self, env_name: str):
        super(EnvVarsSource, self).__init__('',
                                            encoding=sys.getfilesystemencoding())
        self.env_name = env_name  # type: str

    @rethrow_as(KeyError, ConfigurationError)
    def provide(self) -> dict:
        self.root = os.environ[self.env_name]
        return super(EnvVarsSource, self).provide()



from .base import Metric, MetricLevel, LeafMetric, EmbeddedSubmetrics, DISABLED, RUNTIME, \
    DEBUG, INHERIT
from .callable import CallableMetric
from .counter import CounterMetric
from .cps import ClicksPerTimeUnitMetric
from .empty import EmptyMetric
from .histogram import HistogramMetric
from .linkfail import LinkfailMetric
from .registry import register_metric, METRIC_NAMES_TO_CLASSES
from .simple import IntegerMetric, FloatMetric
from .summary import QuantileMetric, SummaryMetric

__all__ = ['Metric', 'LeafMetric', 'EmbeddedSubmetrics', 'ClicksPerTimeUnitMetric',
           'IntegerMetric', 'FloatMetric',
           'QuantileMetric', 'register_metric', 'METRIC_NAMES_TO_CLASSES', 'SummaryMetric',
           'HistogramMetric', 'EmptyMetric', 'LinkfailMetric', 'CallableMetric', 'MetricLevel',
           'INHERIT', 'DEBUG', 'RUNTIME', 'DISABLED']



import collections
import typing as tp

from .sorted_list import SortedList

T = tp.TypeVar('T')


class Ranking(tp.Generic[T]):
    """
    A set of objects with them being ranked by their single property with the assumption that this
    property changes only when we want to.

    Positions are counted from 0, where 0 has the least key value.

    Essentially, this is a SortedList with the option to query at which position can be given
    element found.

    Example usage:

    >>> Entry = collections.namedtuple('Entry', ('key', ))
    >>> e1 = Entry(2)
    >>> e2 = Entry(3)
    >>> e3 = Entry(5)
    >>> ranking = Ranking([e1, e2, e3], lambda e: e.key)
    >>> assert ranking[0] == e1     # Get the first element
    >>> assert ranking[-1] == e3    # Get the last element
    >>> assert ranking.get_position_of(e1) == 0
    """
    __slots__ = ('items', 'key', 'ranking', 'element_to_position')

    def __init__(self, items: tp.List[T], key: tp.Callable[[T], int]):
        self.items = items
        self.key = key
        self.ranking = SortedList(items, key=key)  # type: SortedList[T]
        self.element_to_position = {}  # type: tp.Dict[int, int]
        for position, item in enumerate(self.ranking):
            self.element_to_position[id(item)] = position

    def calculate_ranking_for(self, item: T) -> int:
        return self.element_to_position[id(item)]

    def add(self, item: T) -> None:
        """
        Add a single element to the ranking and recalculate it
        """
        index = self.ranking.add(item)
        for position, item in enumerate(self.ranking[index:], start=index):
            self.element_to_position[id(item)] = position

    def remove(self, item: T) -> None:
        """
        Remove a single element from the ranking and recalculate it
        """
        index = self.ranking.index(item)
        self.ranking.remove(item)
        for position, item in enumerate(self.ranking[index:], start=index):
            self.element_to_position[id(item)] = position

    def get_position_of(self, item: T) -> int:
        """
        Return the position in the ranking of element item

        :param item: element to return the position for
        :return: position
        :raises ValueError: this element is not in the ranking
        """
        return self.ranking.index(item)

    def __getitem__(self, item: int) -> T:
        """
        Return n-th item in ranking.

        :param item: position in ranking. Can be negative, or even a slice
        """
        return self.ranking[item]

    def get_sorted(self) -> tp.Iterator[T]:
        """
        Return all the elements sorted from the least key value to the highest key value
        """
        yield from self.ranking
